require "util";
require "defines";

--Initialise global variables and catch the case where save files with installations of an older version of the mod are introduced to the new functionality.
script.on_init(function()
    if global.shr == nil then
        global.shr = {};
    end

    if global.shr.growing == nil then
        global.shr.growing = {};
    end
        
    --Time in seconds it takes to reach the minimum growth time threshold; after a growing plant survives this number of seconds, it has a random chance of growing into the next stage of plant - see it as "the plant needs to grow AT LEAST this number of seconds".
        global.shr.growTimePlant = 5;
    
    --Time in seconds it takes for a plant in specific grow states to die if it doesn't progress into further stages of adulthood after this much time has elapsed; simulates disease, parasites etc. - some factors that make it harder to grow plants.
    --Make other parts of the mod hook into these variables (it's possible; look it up) to be able to research technologies that increase this amount i.e. lengthen life span of / make trees' lives healthier in general.
        --Causes death timer to be roughly 250% of minimum growth timer.
        global.shr.deathTimePlant = global.shr.growTimePlant + 3*math.sqrt(global.shr.growTimePlant);
    
    --Influences growth and death time calculations. The higher the number, the longer each should take.
        global.shr.growTimePlantScalar = 600; --600
        
    if global.shr.treeType == nil then
        global.shr.treeType = {};
    end
        
    if global.shr.checkAlive == nil then
        global.shr.checkAlive = {};
    end
    
    if global.shr.foundPlant == nil then
        global.shr.foundPlant = {};
    end
        
    if global.shr.newPlant == nil then
        global.shr.newPlant = {};
    end
end)

script.on_load(function()
    if global.shr == nil then
        global.shr = {};
    end

    if global.shr.growing == nil then
        global.shr.growing = {};
    end
        
        global.shr.growTimePlant = 5;
        
        global.shr.deathTimePlant = global.shr.growTimePlant + 3*math.sqrt(global.shr.growTimePlant);
        
        global.shr.growTimePlantScalar = 600; --600
    
    if global.shr.treeType == nil then
        global.shr.treeType = {};
    end    
        
    if global.shr.checkAlive == nil then
        global.shr.checkAlive = {};
    end
        
    if global.shr.foundPlant == nil then
        global.shr.foundPlant = {};
    end
        
    if global.shr.newPlant == nil then
        global.shr.newPlant = {};
    end
end)

--When an entity is placed by the player, define the initialised global variables for that item
script.on_event(defines.events.on_built_entity, function(event)
    if ((event.created_entity.name == "sapling-budding") or (event.created_entity.name == "sapling-young") or (event.created_entity.name == "sapling") or (event.created_entity.name == "sapling-maturing")) then
        --Define the created item by its position as well as its time of creation in game ticks / 60, i.e. seconds, which should, in pairs, be unique.
        global.shr.growing[event.created_entity.position] = math.floor(game.tick / 60);
        global.shr.treeType[event.created_entity.position] = nil;
        global.shr.checkAlive[event.created_entity.position] = nil;
        global.shr.foundPlant[event.created_entity.position] = nil;
        global.shr.newPlant[event.created_entity.position] = nil;
    end
end)

--When an entity is placed by a Construction Robot, define the initialised global variables for that item
--DUPLICATE OF ABOVE FUNCTION



--****               <****----****>              ****--
--****Table printing function stolen from MoMods.****--
--Thank you ludsoe.
local function PrintTab(Table)
	for i,d in pairs(Table) do
		local T = type(d)
		if T == "table" then
			game.player.print("Table: "..i)
			PrintTab(d)
		else
			game.player.print(i.." : "..tostring(d))
		end
	end
end
--****               <****----****>              ****--


local function findEntities(k, areaEntity, nameEntity)
   return game.get_surface(1).find_entities_filtered{area = {{k.x - areaEntity, k.y - areaEntity}, {k.x + areaEntity, k.y + areaEntity}}, name = nameEntity};
end


local function findAndDestroyEntity(k, destroyEntity)
    
    --Find at which growth state the plant was
    --Kill it and let it be known that a plant was killed
    for _, entity in pairs( findEntities(k, 0.03, 'sapling-budding') ) do
        global.shr.foundPlant[k] = entity.name;
        if destroyEntity then
            entity.destroy();
        end
    end

    for _, entity in pairs( findEntities(k, 0.03, 'sapling-young') ) do
        global.shr.foundPlant[k] = entity.name;
        if destroyEntity then
            entity.destroy();
        end
    end

    for _, entity in pairs( findEntities(k, 0.03, 'sapling') ) do
        global.shr.foundPlant[k] = entity.name;
        if destroyEntity then
            entity.destroy();
        end
    end

    for _, entity in pairs( findEntities(k, 0.03, 'sapling-maturing') ) do
        global.shr.foundPlant[k] = entity.name;
        if destroyEntity then
            entity.destroy();
        end
    end
    
    return global.shr.foundPlant[k];
end


script.on_event(defines.events.on_tick, function(event)
    --Every second, for each pair of growing entities (as referenced to by their position in the world) and their time of creation within a tick cycle, do...
    if game.tick % 60 == 0 then
        for k, v in pairs(global.shr.growing) do

            --Check whether the plant is still there (it has not died or been mined/collected by robots)
            global.shr.checkAlive[k] = findAndDestroyEntity(k, false);
            if global.shr.checkAlive[k] == nil then
                --The plant entity no longer exists. Remove all timers and clear the array entry for it.
                global.shr.growing[k] = nil;
            end
            --Throw the value in checkAlive away. We don't need it.
            global.shr.checkAlive[k] = nil;
                
            game.player.print(((game.tick / 60) - (v + global.shr.growTimePlant)) / global.shr.growTimePlantScalar);
            
            game.player.print(((game.tick / 60) - (v + global.shr.growTimePlant)) .." > ".. math.floor(global.shr.deathTimePlant * (global.shr.growTimePlantScalar^(1/4))));
                
            --If a random fraction between 0 and 1 is less than the current time stamp in seconds minus (the time stamp the growing plant was created + its minimum growth time), normalised for seconds such that the amount of seconds the plant has been alive past its minimum growth time causes it to become increasingly more likely that the condition is met...
            if math.random() < ((game.tick / 60) - (v + global.shr.growTimePlant)) / global.shr.growTimePlantScalar then
                    
                findAndDestroyEntity(k, true);
                
                if global.shr.foundPlant[k] then
                    --Advance the killed plant to the next stage in the growth cycle by spawning in the new entity.
                    if global.shr.foundPlant[k] == "sapling-budding" then
                            
                        global.shr.newPlant[k] = game.get_surface(1).create_entity({ name='sapling-young', amount=1, position=k});
                        global.shr.growing[global.shr.newPlant[k].position] = math.floor(game.tick / 60);
                        global.shr.newPlant[k] = nil;
                            
                    elseif global.shr.foundPlant[k] == "sapling-young" then
                            
                        global.shr.newPlant[k] = game.get_surface(1).create_entity({ name='sapling', amount=1, position=k});
                        global.shr.growing[global.shr.newPlant[k].position] = math.floor(game.tick / 60);
                        global.shr.newPlant[k] = nil;
                            
                    elseif global.shr.foundPlant[k] == "sapling" then
                            
                        global.shr.newPlant[k] = game.get_surface(1).create_entity({ name='sapling-maturing', amount=1, position=k});
                        global.shr.growing[global.shr.newPlant[k].position] = math.floor(game.tick / 60);
                        global.shr.newPlant[k] = nil;
                            
                    elseif global.shr.foundPlant[k] == "sapling-maturing" then
                            
                        global.shr.treeType[k] = math.random(6);

                        if global.shr.treeType[k] == 6 then
                            global.shr.treeType[k] = "tree-02-red";
                        else
                            global.shr.treeType[k] = "tree-0"..global.shr.treeType[k];
                        end

                        game.get_surface(1).create_entity({ name=global.shr.treeType[k], amount=1, position=k});
                        global.shr.treeType[k] = nil;
                    end
                        
                    global.shr.foundPlant[k] = nil;
                end
                --Kill the predecessor's entry in the growing array, also deleting its growth timer.
                global.shr.growing[k] = nil;
            
            
            elseif ((game.tick / 60) - (v + global.shr.growTimePlant)) > math.floor(global.shr.deathTimePlant * (global.shr.growTimePlantScalar^(1/4))) then
                --This plant was either removed, or it's taken too long to grow into (further stages of) adulthood. Consider it dead. In fact, let's replace it with a dead tree, if the entity still exists. If it does no longer exist, consider this part of the script clean-up to reduce the amount of infinitely ticking functions that under other circumstances would keep on going forever.
                    
                findAndDestroyEntity(k, true);
                
                if global.shr.foundPlant[k] ~= nil then    
                    game.get_surface(1).create_entity({ name="dry-young-tree", amount=1, position=k});
                end  
                
                global.shr.foundPlant[k] = nil;
                --Stop the plant's growth timer and remove its entry in the array completely.
                global.shr.growing[k] = nil;
            end
                
            global.shr.foundPlant[k] = nil;
        end
    end
end)